-- vim: ts=2 sw=2 ai et enc=utf8 fenc=utf8

local RC = RaidChecklist

if not RC then return end

-- Map from GUID to name
RC.provider_names = {}

-- Map from (de)buff key to list of GUIDs currently providing said de(buff)
RC.providers = {}
for i,b in ipairs (RC.buffs) do RC.providers[b] = {} end
for i,b in ipairs (RC.debuffs) do RC.providers[b] = {} end
for i,b in ipairs (RC.dispels) do RC.providers[b] = {} end
for i,b in ipairs (RC.ccs) do RC.providers[b] = {} end
for i,b in ipairs (RC.miscs) do RC.providers[b] = {} end

local function remove_provider (rguid)
  -- Iterate over all the (de)buffs
  for key,guidlist in pairs (RC.providers) do
    guidlist[rguid] = nil
  end
  if (RC.OnRefresh) then RC:OnRefresh () end
end

local function add_provider (buff, guid)
  RC.providers[buff][guid] = true
  if (RC.OnRefresh) then RC:OnRefresh () end
end

-- Works out which spec is used based on talent point allocations.
-- Returns 0 if no main tree is spec'd into, else 1-3 indicating tree.
local function which_spec (n1, n2, n3)
  return (n1 > n2 and n1 > n3 and 1) or (n2 > n1 and n2 > n3 and 2) or (n3 > n1 and n3 > n2 and 3) or 0
end

local B = RC.buffs
local C = RC.ccs
local D = RC.debuffs
local S = RC.dispels
local M = RC.miscs

-- Matrix based on select(2, UnitBaseClass(unit)) and spec (0-4) to a list of (de)buffs provided
-- This is almost a duplication of the "availablefrom" table over in raid_checklist, but that one
-- has to be human readable and there was just no neat way to satisfy both in a reasonable manner :(
local matrix = {
  WARRIOR = {
    [0] = { B.AGISTR, B.STAM, M.SNARE, M.DISARM, M.INTERRUPT }, -- All warriors can provide these even without a spec
    [1] = { B.AGISTR, B.STAM, M.SNARE, M.DISARM, M.INTERRUPT, D.PHYSDMGTAKEN, D.BLEEDDMGTAKEN, D.MORTALSTRIKE }, -- Arms
    [2] = { B.AGISTR, B.STAM, M.SNARE, M.DISARM, M.INTERRUPT, B.CRIT, D.MORTALSTRIKE }, -- Fury
    [3] = { B.AGISTR, B.STAM, M.SNARE, M.DISARM, M.INTERRUPT, D.DEMOROAR, D.ATTACKSPD, D.SUNDER }, -- Protection
  },
  HUNTER = {
    [0] = { D.MORTALSTRIKE, M.SNARE, S.ENRAGE, S.OFFMAGIC, C.CCANY, C.FEARBEASTS },
    [1] = { D.MORTALSTRIKE, M.SNARE, S.ENRAGE, S.OFFMAGIC, C.CCANY, C.FEARBEASTS, B.ALLDMG }, -- BM
    [2] = { D.MORTALSTRIKE, M.SNARE, S.ENRAGE, S.OFFMAGIC, C.CCANY, C.FEARBEASTS, B.ATTACKPWR, M.INTERRUPT }, -- Marksman
    [3] = { D.MORTALSTRIKE, M.SNARE, S.ENRAGE, S.OFFMAGIC, C.CCANY, C.FEARBEASTS, B.PHYSHASTE }, -- Survival
  },
  DEATHKNIGHT = {
    [0] = { B.AGISTR, D.ATTACKSPD, M.SNARE, D.CASTSPD, M.INTERRUPT },
    [1] = { B.AGISTR, D.ATTACKSPD, M.SNARE, D.CASTSPD, M.INTERRUPT, B.ATTACKPWR, D.DEMOROAR }, -- Blood
    [2] = { B.AGISTR, D.ATTACKSPD, M.SNARE, D.CASTSPD, M.INTERRUPT, B.PHYSHASTE, D.PHYSDMGTAKEN }, -- Frost
    [3] = { B.AGISTR, D.ATTACKSPD, M.SNARE, D.CASTSPD, M.INTERRUPT, D.SPELLDMGTAKEN }, -- Unholy
  },
  PRIEST = {
    [0] = { B.STAM, B.BURSTMANA, S.DEFMAGIC, S.OFFMAGIC, C.CCUNDEAD },
    [1] = { B.STAM, B.BURSTMANA, S.DEFMAGIC, S.OFFMAGIC, C.CCUNDEAD, S.DISEASE }, -- Discipline
    [2] = { B.STAM, B.BURSTMANA, S.DEFMAGIC, S.OFFMAGIC, C.CCUNDEAD, S.DISEASE }, -- Holy
    [3] = { B.STAM, B.BURSTMANA, S.DEFMAGIC, S.OFFMAGIC, C.CCUNDEAD, B.SPELLHASTE, B.REPLENISHMENT, D.MORTALSTRIKE, M.SNARE, M.DISARM }, -- Shadow
  },
  SHAMAN = {
    [0] = { B.AGISTR, B.SPELLPWR6, B.PHYSHASTE, B.SPELLHASTE, B.BLOODLUST, B.MANAREGEN, M.SNARE, B.ARMOR, B.PUSHBACKRES, D.ATTACKSPD, S.OFFMAGIC, S.CURSE, C.CCBEASTS, C.CCELEMENTALS, C.CCHUMANOIDS, M.INTERRUPT },
    [1] = { B.AGISTR, B.SPELLPWR6, B.PHYSHASTE, B.SPELLHASTE, B.BLOODLUST, B.MANAREGEN, M.SNARE, B.ARMOR, B.PUSHBACKRES, D.ATTACKSPD, S.OFFMAGIC, S.CURSE, C.CCBEASTS, C.CCELEMENTALS, C.CCHUMANOIDS, M.INTERRUPT, B.CRIT, B.SPELLPWR10, M.KNOCKBACK }, -- Elemental
    [2] = { B.AGISTR, B.SPELLPWR6, B.PHYSHASTE, B.SPELLHASTE, B.BLOODLUST, B.MANAREGEN, M.SNARE, B.ARMOR, B.PUSHBACKRES, D.ATTACKSPD, S.OFFMAGIC, S.CURSE, C.CCBEASTS, C.CCELEMENTALS, C.CCHUMANOIDS, M.INTERRUPT, B.ATTACKPWR }, -- Enhancement
    [3] = { B.AGISTR, B.SPELLPWR6, B.PHYSHASTE, B.SPELLHASTE, B.BLOODLUST, B.MANAREGEN, M.SNARE, B.ARMOR, B.PUSHBACKRES, D.ATTACKSPD, S.OFFMAGIC, S.CURSE, C.CCBEASTS, C.CCELEMENTALS, C.CCHUMANOIDS, M.INTERRUPT, B.BURSTMANA, S.DEFMAGIC }, -- Restoration
  },
  PALADIN = {
    [0] = { B.STATS, B.ATTACKPWR, B.MANAREGEN, B.ARMOR, B.PUSHBACKRES, S.DISEASE, S.POISON, C.FEARUNDEAD },
    [1] = { B.STATS, B.ATTACKPWR, B.MANAREGEN, B.ARMOR, B.PUSHBACKRES, S.DISEASE, S.POISON, C.FEARUNDEAD, S.DEFMAGIC }, -- Holy
    [2] = { B.STATS, B.ATTACKPWR, B.MANAREGEN, B.ARMOR, B.PUSHBACKRES, S.DISEASE, S.POISON, C.FEARUNDEAD, D.ATTACKSPD, D.DEMOROAR }, -- Protection
    [3] = { B.STATS, B.ATTACKPWR, B.MANAREGEN, B.ARMOR, B.PUSHBACKRES, S.DISEASE, S.POISON, C.FEARUNDEAD, B.REPLENISHMENT, B.ALLDMG, C.CCDEMONS, C.CCDRAGONKIN, C.CCGIANTS, C.CCHUMANOIDS, C.CCUNDEAD }, -- Retribution
  },
  DRUID = {
    [0] = { B.STATS },
    [1] = { B.STATS, S.POISON, S.CURSE, S.ENRAGE, M.ROOT, C.CCBEASTS, C.CCDRAGONKIN, B.BURSTMANA, B.SPELLHASTE, D.SPELLDMGTAKEN, M.KNOCKBACK }, -- Balance
    [2] = { B.STATS, B.CRIT, D.BLEEDDMGTAKEN, D.ATTACKSPD, D.DEMOROAR, D.SUNDER, M.SNARE }, -- Feral (demo/sunder only in bear though)
    [3] = { B.STATS, S.POISON, S.CURSE, S.ENRAGE, M.ROOT, C.CCBEASTS, C.CCDRAGONKIN, B.BURSTMANA, B.REPLENISHMENT, S.DEFMAGIC }, -- Restoration
  },
  WARLOCK = {
    [0] = { D.SPELLDMGTAKEN, D.CASTSPD, D.DEMOROAR, C.CCDEMONS, C.CCELEMENTALS, C.FEARANY },
    [1] = { D.SPELLDMGTAKEN, D.CASTSPD, D.DEMOROAR, C.CCDEMONS, C.CCELEMENTALS, C.FEARANY, M.SNARE }, -- Affliction
    [2] = { D.SPELLDMGTAKEN, D.CASTSPD, D.DEMOROAR, C.CCDEMONS, C.CCELEMENTALS, C.FEARANY, B.SPELLPWR10  }, -- Demonology
    [3] = { D.SPELLDMGTAKEN, D.CASTSPD, D.DEMOROAR, C.CCDEMONS, C.CCELEMENTALS, C.FEARANY, B.REPLENISHMENT, D.SPELLCRIT }, -- Destruction (spell crit accessible from other trees though "Shadow and Flame")
  },
  ROGUE = {
    [0] = { D.CASTSPD, D.MORTALSTRIKE, D.SUNDER, M.SNARE, M.DISARM, S.ENRAGE, C.CCBEASTS, C.CCDEMONS, C.CCDRAGONKIN, C.CCHUMANOIDS, M.INTERRUPT },
    [1] = { D.CASTSPD, D.MORTALSTRIKE, D.SUNDER, M.SNARE, M.DISARM, S.ENRAGE, C.CCBEASTS, C.CCDEMONS, C.CCDRAGONKIN, C.CCHUMANOIDS, M.INTERRUPT, D.SPELLDMGTAKEN }, -- Assasination
    [2] = { D.CASTSPD, D.MORTALSTRIKE, D.SUNDER, M.SNARE, M.DISARM, S.ENRAGE, C.CCBEASTS, C.CCDEMONS, C.CCDRAGONKIN, C.CCHUMANOIDS, M.INTERRUPT, D.PHYSDMGTAKEN }, -- Combat
    [3] = { D.CASTSPD, D.MORTALSTRIKE, D.SUNDER, M.SNARE, M.DISARM, S.ENRAGE, C.CCBEASTS, C.CCDEMONS, C.CCDRAGONKIN, C.CCHUMANOIDS, M.INTERRUPT, B.CRIT, D.BLEEDDMGTAKEN }, -- Subtlety
  },
  MAGE = {
    [0] = { B.SPELLPWR6, B.BLOODLUST, B.MANA, C.CCBEASTS, C.CCHUMANOIDS, M.INTERRUPT, M.SNARE, S.CURSE },
    [1] = { B.SPELLPWR6, B.BLOODLUST, B.MANA, C.CCBEASTS, C.CCHUMANOIDS, M.INTERRUPT, M.SNARE, S.CURSE, D.CASTSPD, B.ALLDMG, D.MORTALSTRIKE }, -- Arcane
    [2] = { B.SPELLPWR6, B.BLOODLUST, B.MANA, C.CCBEASTS, C.CCHUMANOIDS, M.INTERRUPT, M.SNARE, S.CURSE, D.SPELLCRIT }, -- Fire
    [3] = { B.SPELLPWR6, B.BLOODLUST, B.MANA, C.CCBEASTS, C.CCHUMANOIDS, M.INTERRUPT, M.SNARE, S.CURSE, B.REPLENISHMENT }, -- Frost
  }
}

local function provides (class, spec)
  return matrix[class][spec]
end

RC.OnTalentUpdate = function (self, event, guid, unit, spec, n1, n2, n3)
  -- Update the talent status
  RC.missing_talents_for = RC.gt:GetTalentMissingNames ()

--if event == "LibGroupTalents_Update" then DEFAULT_CHAT_FRAME:AddMessage ("RC: updated talents for GUID "..(guid or "nil")..", unit "..unit) end

  -- Clear any previous entries for this guid
  remove_provider (guid)

  local spec = which_spec (n1, n2, n3)
  local class = select (2, UnitClassBase (unit))
  local list = matrix[class][spec]
  for i,key in ipairs (list) do
    add_provider (key, guid)
  end
end

RC.OnTalentAdd = function (self, event, guid, unit, name, realm)
  local fullname = (realm and name.."-"..realm) or name
  RC.provider_names[guid] = ClassColorString (select (2,UnitClassBase (unit)))..fullname.."|r"

  -- Update based on class alone to start off with. Better than waiting until in inspect range!
  RC:OnTalentUpdate ("fake", guid, unit, 0, 0, 0, 0)

  --RaidChecklistButton:UNIT_PET(unit) -- We are probably better off waiting for a proper UNIT_PET event I think

  --DEFAULT_CHAT_FRAME:AddMessage ("RC: discovered provider GUID "..(guid or "nil")..", named "..fullname)
end

RC.OnTalentRemove = function (self, event, guid, name, realm)
  -- Update the talent status
  RC.missing_talents_for = RC.gt:GetTalentMissingNames ()

  remove_provider (guid)
  RC.provider_names[guid] = nil

  --DEFAULT_CHAT_FRAME:AddMessage ("RC: removed provider GUID "..(guid or "nil"))
end

RC.gt = LibStub ("LibGroupTalents-1.0")
RC.gt.RegisterCallback (RC, "LibGroupTalents_Update", "OnTalentUpdate")
RC.gt.RegisterCallback (RC, "LibGroupTalents_Add", "OnTalentAdd")
RC.gt.RegisterCallback (RC, "LibGroupTalents_Remove", "OnTalentRemove")

